from __future__ import annotations

from unittest.mock import MagicMock

from sphinx.application import Sphinx

from sphinx_autodoc_typehints._formats import detect_format
from sphinx_autodoc_typehints._formats._base import InsertIndexInfo
from sphinx_autodoc_typehints._formats._numpydoc import NumpydocFormat, _convert_numpydoc_to_sphinx_fields
from sphinx_autodoc_typehints._formats._sphinx import SphinxFieldListFormat


def test_convert_parameters_section() -> None:
    lines = [
        "",
        "Summary.",
        "",
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
        "    **y** : str",
        "        The y value.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param x: The x value." in lines
    assert ":type x: int" in lines
    assert ":param y: The y value." in lines
    assert ":type y: str" in lines
    assert ":Parameters:" not in lines


def test_convert_returns_single() -> None:
    lines = [
        ":Returns:",
        "",
        "    **result** : str",
        "        The combined result.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":returns: The combined result." in lines
    assert ":Returns:" not in lines


def test_convert_returns_no_name() -> None:
    lines = [
        ":Returns:",
        "",
        "    str",
        "        A greeting.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":returns: A greeting." in lines


def test_convert_returns_multiple() -> None:
    lines = [
        ":Returns:",
        "",
        "    **name** : str",
        "        The name.",
        "",
        "    **value** : int",
        "        The value.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":returns:" in lines
    assert any("**name**" in line and "str" in line for line in lines)
    assert any("**value**" in line and "int" in line for line in lines)


def test_convert_raises() -> None:
    lines = [
        ":Raises:",
        "",
        "    ValueError",
        "        If x is negative.",
        "",
        "    TypeError",
        "        If x is not an integer.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":raises ValueError: If x is negative." in lines
    assert ":raises TypeError: If x is not an integer." in lines


def test_convert_yields_single() -> None:
    lines = [
        ":Yields:",
        "",
        "    int",
        "        A number.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":Yields: A number." in lines


def test_convert_yields_multiple() -> None:
    lines = [
        ":Yields:",
        "",
        "    **name** : str",
        "        The name.",
        "",
        "    **value** : int",
        "        The value.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":Yields:" in lines
    assert any("**name**" in line and "str" in line for line in lines)
    assert any("**value**" in line and "int" in line for line in lines)


def test_convert_other_parameters() -> None:
    lines = [
        ":Other Parameters:",
        "",
        "    **flag** : bool",
        "        A flag.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param flag: A flag." in lines
    assert ":type flag: bool" in lines


def test_no_numpydoc_content_is_noop() -> None:
    original = [
        "",
        "A normal docstring.",
        "",
        ":param x: The x value.",
        ":type x: int",
    ]
    lines = list(original)
    _convert_numpydoc_to_sphinx_fields(lines)
    assert lines == original


def test_multiline_description() -> None:
    lines = [
        ":Parameters:",
        "",
        "    **data** : dict",
        "        The data dictionary.",
        "        It can span multiple lines.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param data: The data dictionary. It can span multiple lines." in lines
    assert ":type data: dict" in lines


def test_param_without_type() -> None:
    lines = [
        ":Parameters:",
        "",
        "    **x**",
        "        The x value.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param x: The x value." in lines
    assert not any(line.startswith(":type x:") for line in lines)


def test_param_without_description() -> None:
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param x:" in lines
    assert ":type x: int" in lines


def test_raises_with_named_exception() -> None:
    lines = [
        ":Raises:",
        "",
        "    **err** : ValueError",
        "        If something is wrong.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":raises err: If something is wrong." in lines


def test_returns_single_no_description() -> None:
    lines = [
        ":Returns:",
        "",
        "    **result** : str",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert not any(line.startswith(":returns:") for line in lines)


def test_multiple_sections() -> None:
    lines = [
        "Summary.",
        "",
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
        ":Returns:",
        "",
        "    **result** : str",
        "        The result.",
        "",
        ":Raises:",
        "",
        "    ValueError",
        "        If bad.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param x: The x value." in lines
    assert ":type x: int" in lines
    assert ":returns: The result." in lines
    assert ":raises ValueError: If bad." in lines


def test_numpydoc_format_detect() -> None:
    assert NumpydocFormat.detect([":Parameters:", "    **x** : int", "        desc"])
    assert NumpydocFormat.detect([":Returns:", "    str", "        result"])
    assert not NumpydocFormat.detect([":param x: foo", ":type x: int"])


def test_numpydoc_format_find_param() -> None:
    fmt = NumpydocFormat()
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
    ]
    idx = fmt.find_param(lines, "x")
    assert idx is not None
    assert lines[idx].startswith(":param x:")


def test_numpydoc_format_find_preexisting_type() -> None:
    fmt = NumpydocFormat()
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
    ]
    fmt._ensure_converted(lines)  # noqa: SLF001
    annotation, found = fmt.find_preexisting_type(lines, "x")
    assert found
    assert "int" in annotation


def test_numpydoc_format_add_undocumented_param() -> None:
    fmt = NumpydocFormat()
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
    ]
    idx = fmt.add_undocumented_param(lines, "y")
    assert idx is not None
    assert ":param y:" in lines[idx]


def test_numpydoc_format_inject_param_type() -> None:
    fmt = NumpydocFormat()
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
    ]
    fmt._ensure_converted(lines)  # noqa: SLF001
    idx = fmt.find_param(lines, "x")
    assert idx is not None
    fmt.inject_param_type(lines, "x", "int", idx + 1)
    assert ":type x: int" in lines


def test_numpydoc_format_get_arg_name_from_line() -> None:
    fmt = NumpydocFormat()
    assert fmt.get_arg_name_from_line(":param x: foo") == "x"
    assert fmt.get_arg_name_from_line("not a param line") is None


def test_numpydoc_format_append_default() -> None:
    """Line 199: append_default delegates to SphinxFieldListFormat."""
    fmt = NumpydocFormat()
    lines = [":param x: the x"]
    result = fmt.append_default(lines, 0, ":type x: int", " (default: ``0``)", after=False)
    assert result == ":type x: int (default: ``0``)"


def test_convert_section_without_blank_line_after_header() -> None:
    """Branch 99->102: no empty line between section header and first entry."""
    lines = [
        ":Parameters:",
        "    **x** : int",
        "        The x value.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param x: The x value." in lines
    assert ":type x: int" in lines


def test_convert_yields_single_no_description_then_params() -> None:
    """Branch 133->90: single Yields entry without description followed by another section."""
    lines = [
        ":Yields:",
        "",
        "    **result** : int",
        "",
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x.",
        "",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert not any(line.startswith(":Yields:") for line in lines)
    assert ":param x: The x." in lines
    assert ":type x: int" in lines


def test_detect_format_returns_numpydoc_for_numpydoc_lines() -> None:
    lines = [":Parameters:", "", "    **x** : int", "        desc"]
    fmt = detect_format(lines)
    assert isinstance(fmt, NumpydocFormat)


def test_detect_format_returns_sphinx_for_plain_lines() -> None:
    lines = [":param x: The x value."]
    fmt = detect_format(lines)
    assert isinstance(fmt, SphinxFieldListFormat)


def test_parse_numpydoc_entries_breaks_on_unrecognized_format() -> None:
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "not an entry",
    ]
    _convert_numpydoc_to_sphinx_fields(lines)
    assert ":param x: The x value." in lines


def test_numpydoc_format_get_rtype_insert_info() -> None:
    fmt = NumpydocFormat()
    lines = [
        ":Parameters:",
        "",
        "    **x** : int",
        "        The x value.",
        "",
    ]
    app: Sphinx = MagicMock(spec=Sphinx)
    app.env = MagicMock()
    result = fmt.get_rtype_insert_info(app, lines)
    assert result is not None


def test_numpydoc_format_inject_rtype() -> None:
    fmt = NumpydocFormat()
    lines = [":param x: value"]
    info = InsertIndexInfo(insert_index=len(lines))
    fmt.inject_rtype(lines, ":py:class:`str`", info, use_rtype=True)
    assert any("rtype" in line for line in lines)



from __future__ import annotations

import re
import sys
from pathlib import Path
from textwrap import dedent, indent
from typing import TYPE_CHECKING, Any, Literal, NewType, TypeVar

import pytest
from conftest import normalize_sphinx_text

if TYPE_CHECKING:
    from collections.abc import Callable
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

T = TypeVar("T")
W = NewType("W", str)


def expected(expected: str, **options: Any) -> Callable[[T], T]:
    def dec(val: T) -> T:
        val.EXPECTED = expected  # ty: ignore[unresolved-attribute]
        val.OPTIONS = options  # ty: ignore[unresolved-attribute]
        return val

    return dec


def warns(pattern: str) -> Callable[[T], T]:
    def dec(val: T) -> T:
        val.WARNING = pattern  # ty: ignore[unresolved-attribute]
        return val

    return dec


ArrayLike = Literal["test"]


class _SchemaMeta(type):  # noqa: PLW1641
    def __eq__(cls, other: object) -> bool:
        return True


class Schema(metaclass=_SchemaMeta):
    pass


@expected(
    """
mod.f(s)

   Do something.

   Parameters:
      **s** ("Schema") -- Some schema.

   Return type:
      "Schema"
"""
)
def f(s: Schema) -> Schema:
    """
    Do something.

    Args:
        s: Some schema.
    """
    return s


class AliasedClass: ...


@expected(
    """
mod.g(s)

   Do something.

   Parameters:
      **s** ("Class Alias") -- Some schema.

   Return type:
      "Class Alias"
"""
)
def g(s: AliasedClass) -> AliasedClass:
    """
    Do something.

    Args:
        s: Some schema.
    """
    return s


@expected(
    f"""\
mod.function(x, y)

   Function docstring.

   Parameters:
      * **x** ({'Array | "None"' if sys.version_info >= (3, 14) else '"Optional"[Array]'}) -- foo

      * **y** ("Schema") -- boo

   Returns:
      something

   Return type:
      bytes

""",
)
def function(x: ArrayLike | None, y: Schema) -> str:
    """
    Function docstring.

    :param x: foo
    :param y: boo
    :return: something
    :rtype: bytes
    """


# Config settings for each test run.
# Config Name: Sphinx Options as Dict.
configs = {"default_conf": {"autodoc_type_aliases": {"ArrayLike": "Array", "AliasedClass": '"Class Alias"'}}}


@pytest.mark.parametrize("val", [x for x in globals().values() if hasattr(x, "EXPECTED")])
@pytest.mark.parametrize("conf_run", list(configs.keys()))
@pytest.mark.sphinx("text", testroot="integration")
def test_integration(
    app: SphinxTestApp, status: StringIO, warning: StringIO, monkeypatch: pytest.MonkeyPatch, val: Any, conf_run: str
) -> None:
    template = ".. autofunction:: mod.{}"

    (Path(app.srcdir) / "index.rst").write_text(template.format(val.__name__))
    app.config.__dict__.update(configs[conf_run])
    app.config.__dict__.update(val.OPTIONS)
    monkeypatch.setitem(sys.modules, "mod", sys.modules[__name__])
    app.build()
    assert "build succeeded" in status.getvalue()  # Build succeeded

    regexp = getattr(val, "WARNING", None)
    value = warning.getvalue().strip()
    if regexp:
        msg = f"Regex pattern did not match.\n Regex: {regexp!r}\n Input: {value!r}"
        assert re.search(regexp, value), msg
    elif not re.search(r"WARNING: Inline strong start-string without end-string.", value):
        assert not value

    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())

    expected = normalize_sphinx_text(val.EXPECTED)
    try:
        assert result.strip() == dedent(expected).strip()
    except Exception:
        indented = indent(f'"""\n{result}\n"""', " " * 4)
        print(f"@expected(\n{indented}\n)\n")  # noqa: T201
        raise



from __future__ import annotations

from textwrap import dedent
from unittest.mock import MagicMock, patch

from sphinx_autodoc_typehints._resolver._type_comments import (
    _normalize_source_lines,
    _split_type_comment_args,
    backfill_type_hints,
)


def test__normalize_source_lines_async_def() -> None:
    source = """
    async def async_function():
        class InnerClass:
            def __init__(self): ...
    """

    expected = """
    async def async_function():
        class InnerClass:
            def __init__(self): ...
    """

    assert _normalize_source_lines(dedent(source)) == dedent(expected)


def test__normalize_source_lines_def_starting_decorator_parameter() -> None:
    source = """
    @_with_parameters(
        _Parameter("self", _Parameter.POSITIONAL_OR_KEYWORD),
        *_proxy_instantiation_parameters,
        _project_id,
        _Parameter(
            "node_numbers",
            _Parameter.POSITIONAL_OR_KEYWORD,
            default=None,
            annotation=Optional[Iterable[int]],
        ),
    )
    def __init__(bound_args):  # noqa: N805
        ...
    """

    expected = """
    @_with_parameters(
        _Parameter("self", _Parameter.POSITIONAL_OR_KEYWORD),
        *_proxy_instantiation_parameters,
        _project_id,
        _Parameter(
            "node_numbers",
            _Parameter.POSITIONAL_OR_KEYWORD,
            default=None,
            annotation=Optional[Iterable[int]],
        ),
    )
    def __init__(bound_args):  # noqa: N805
        ...
    """

    assert _normalize_source_lines(dedent(source)) == dedent(expected)


def test_syntax_error_backfill() -> None:
    def func(x):  # noqa: ANN202
        ...

    backfill_type_hints(func, "func")


def test__normalize_source_lines_no_def() -> None:
    source = "x = 1\ny = 2\n"
    assert _normalize_source_lines(source) == source


def test__split_type_comment_args_empty() -> None:
    assert _split_type_comment_args("") == []


def test__split_type_comment_args_single() -> None:
    assert _split_type_comment_args("int") == ["int"]


def test__split_type_comment_args_multiple() -> None:
    assert _split_type_comment_args("int, str, bool") == ["int", "str", "bool"]


def test__split_type_comment_args_nested_brackets() -> None:
    assert _split_type_comment_args("List[int], Dict[str, int]") == ["List[int]", "Dict[str, int]"]


def test__split_type_comment_args_strips_stars() -> None:
    assert _split_type_comment_args("*args, **kwargs") == ["args", "kwargs"]


def test_backfill_type_hints_no_type_comment_attr() -> None:
    with patch("sphinx_autodoc_typehints._resolver._type_comments.inspect.getsource", return_value="import os\n"):
        assert backfill_type_hints(lambda: None, "test") == {}


def test_backfill_type_hints_multi_child_ast() -> None:
    with patch(
        "sphinx_autodoc_typehints._resolver._type_comments.inspect.getsource",
        return_value="x = 1\ndef foo(): pass\n",
    ):
        assert backfill_type_hints(lambda: None, "test") == {}


def test_backfill_type_hints_with_type_comment() -> None:
    result = backfill_type_hints(_typed_func, "_typed_func")
    assert result == {"x": "int", "y": "str", "return": "bool"}


def test_backfill_type_hints_type_comment_self_insertion() -> None:
    result = backfill_type_hints(_BackfillClass.method, "_BackfillClass.method")
    assert result == {"x": "int", "return": "str"}


def test_backfill_type_hints_type_comment_mismatch() -> None:
    result = backfill_type_hints(_mismatched_type_comment, "_mismatched_type_comment")
    assert result == {"return": "bool"}


def test_backfill_type_hints_posonly_args() -> None:
    result = backfill_type_hints(_posonly_typed_func, "_posonly_typed_func")
    assert result == {"x": "int", "y": "str", "return": "bool"}


def test_backfill_type_hints_empty_return() -> None:
    with patch(
        "sphinx_autodoc_typehints._resolver._type_comments.inspect.getsource",
        return_value="def f(x):\n    # type: (int) -> \n    pass\n",
    ):

        def f(x): ...  # noqa: ANN202

        result = backfill_type_hints(f, "f")
    assert result == {"x": "int"}


def test_backfill_type_hints_unparseable_type_comment() -> None:
    with patch(
        "sphinx_autodoc_typehints._resolver._type_comments.inspect.getsource",
        return_value="def f(x):\n    # type: bad_comment\n    pass\n",
    ):

        def f(x): ...  # noqa: ANN202

        result = backfill_type_hints(f, "f")
    assert result == {}


def test_backfill_type_hints_inline_type_comments() -> None:
    result = backfill_type_hints(_inline_typed_func, "_inline_typed_func")
    assert result == {"x": "int", "y": "str", "return": "bool"}


def test_multi_child_ast_warning_includes_location() -> None:
    mock_logger = MagicMock()
    with (
        patch(
            "sphinx_autodoc_typehints._resolver._type_comments.inspect.getsource",
            return_value="x = 1\ndef foo(): pass\n",
        ),
        patch("sphinx_autodoc_typehints._resolver._type_comments._LOGGER", mock_logger),
    ):
        backfill_type_hints(lambda: None, "test")
    mock_logger.warning.assert_called_once()
    kwargs = mock_logger.warning.call_args.kwargs
    assert "location" in kwargs


# --- module-level fixtures for inspect.getsource ---


class _BackfillClass:
    def method(self, x):  # noqa: ANN202
        # type: (int) -> str
        ...


def _typed_func(x, y):  # noqa: ANN202
    # type: (int, str) -> bool
    ...


def _posonly_typed_func(x, y, /):  # noqa: ANN202
    # type: (int, str) -> bool
    ...


def _inline_typed_func(  # noqa: ANN202
    x,  # type: int
    y,  # type: str
):
    # type: (...) -> bool
    ...


def _mismatched_type_comment(x, y, z):  # noqa: ANN202
    # type: (int) -> bool
    ...
